import sys
from minigrammar import *

class CSVSettings(LanguageSettings):
    @classmethod
    def ignore_characters(cls, char):
        return char == ' ' or char == '\t' or char == '\r'


@repeating("Record", None, None, '\n', False, False)
class Document(CSVSettings):
    def __repr__(self):
        return self.elems.__repr__()


@repeating("Field", None, None, ',', False, False)
class Record(CSVSettings):
    def __repr__(self):
        return self.elems.__repr__()


@regex_pattern(r'^[a-zA-Z0-9_]+$')
class Field(CSVSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


if __name__ == "__main__":
    with open(sys.argv[1], 'r') as file:
        iterator = FileParserIterator(file)
        document = Document(iterator)
        print(document)


import unittest

from minigrammar.iterators import StringParserIterator
from minigrammar.parsing import *
from minigrammar.exceptions import *
from minigrammar.settings import *


class ParsingRegexPatterns(unittest.TestCase):

    def test_case_digest_everything_successfully(self):
        @regex_pattern(r'\b-?\d+\b')
        class Integer(LanguageSettings): pass

        iterator = StringParserIterator("10")
        Integer(iterator)
        self.assertEqual(None, iterator.peek())

    def test_case_regex_mismatch_left_the_iterator_untouched(self):
        @regex_pattern(r'\b-?\d+\b')
        class Integer(LanguageSettings): pass

        iterator = StringParserIterator("XX")
        with self.assertRaises(CannotParseException):
            Integer(iterator)
        self.assertTrue("X" == iterator.peek())

    def test_case_regex_mismatch_on_empty_string_left_the_iterator_untouched(self):
        @regex_pattern(r'\b-?\d+\b')
        class Integer(LanguageSettings): pass

        iterator = StringParserIterator("")
        with self.assertRaises(CannotParseException):
            Integer(iterator)
        self.assertTrue(None == iterator.peek())

    def test_case_regex_ok_on_empty_string_left_the_iterator_untouched(self):
        @regex_pattern(r'')
        class Empty(LanguageSettings): pass

        iterator = StringParserIterator("")
        Empty(iterator)
        self.assertTrue(None == iterator.peek())

    def test_case_regex_for_string_literals(self):
        @regex_pattern(r'^"(?:[^"\\]|\\.)*"$')
        class StringLiteral(LanguageSettings): pass

        iterator = StringParserIterator('"Hello \\" World"X')
        StringLiteral(iterator)
        self.assertEqual("X", iterator.peek())


if __name__ == '__main__':
    unittest.main()



import unittest

from minigrammar.iterators import StringParserIterator
from minigrammar.parsing import *
from minigrammar.settings import *
from minigrammar.exceptions import *


class ParsingExactMatches(unittest.TestCase):

    def test_case_digest_everything_successfully(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        iterator = StringParserIterator("10")
        Number10(iterator)
        self.assertEqual(None, iterator.peek())

    def test_case_nothing_to_digest_exception(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        with self.assertRaises(CannotParseException):
            Number10(StringParserIterator(""))

    def test_case_nothing_to_digest_but_ok(self):
        @exact_match("")
        class Empty(LanguageSettings): pass

        iterator = StringParserIterator("")
        Empty(iterator)
        self.assertEqual(None, iterator.peek())

    def test_case_mismatch_left_the_iterator_untouched(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        iterator = StringParserIterator("50")
        with self.assertRaises(CannotParseException):
            Number10(iterator)
        self.assertEqual("5", iterator.peek())


if __name__ == '__main__':
    unittest.main()



import unittest

from minigrammar.iterators import StringParserIterator
from minigrammar.settings import LanguageSettings
from minigrammar.parsing import *
from minigrammar.exceptions import *


class ParsingEither(unittest.TestCase):

    def test_case_digest_everything_successfully(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @exact_match("Hello")
        class Hello(LanguageSettings): pass

        @either([Hello, Number10])
        class Number10OrHello(LanguageSettings): pass

        iterator = StringParserIterator("Hello")
        Number10OrHello(iterator)
        self.assertEqual(None, iterator.peek())

    def test_case_digest_everything_successfully2(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @exact_match("Hello")
        class Hello(LanguageSettings): pass

        @either([Hello, Number10])
        class Number10OrHello(LanguageSettings): pass

        iterator = StringParserIterator("10")
        Number10OrHello(iterator)
        self.assertEqual(None, iterator.peek())

    def test_case_no_case_matches(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @exact_match("Hello")
        class Hello(LanguageSettings): pass

        @either([Hello, Number10])
        class Number10OrHello(LanguageSettings): pass

        iterator = StringParserIterator("XX")
        with self.assertRaises(CannotParseException):
            Number10OrHello(iterator)
        self.assertEqual("X", iterator.peek())


if __name__ == '__main__':
    unittest.main()



import unittest

from minigrammar.iterators import StringParserIterator
from minigrammar.settings import LanguageSettings
from minigrammar.parsing import *
from minigrammar.exceptions import *


class ParsingChain(unittest.TestCase):

    def test_case_digest_everything_successfully(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @exact_match("Hello")
        class Hello(LanguageSettings): pass

        @chain([Number10, Hello])
        class Number10ThenHello(LanguageSettings): pass

        iterator = StringParserIterator("10Hello")
        Number10ThenHello(iterator)
        self.assertEqual(None, iterator.peek())

    def test_case_full_mismatch(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @exact_match("Hello")
        class Hello(LanguageSettings): pass

        @chain([Number10, Hello])
        class Number10ThenHello(LanguageSettings): pass

        iterator = StringParserIterator("XX")
        with self.assertRaises(CannotParseException):
            Number10ThenHello(iterator)
        self.assertEqual("X", iterator.peek())

    def test_case_full_mismatch_on_empty_string(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @exact_match("Hello")
        class Hello(LanguageSettings): pass

        @chain([Number10, Hello])
        class Number10ThenHello(LanguageSettings): pass

        iterator = StringParserIterator("")
        with self.assertRaises(CannotParseException):
            Number10ThenHello(iterator)
        self.assertEqual(None, iterator.peek())

    def test_case_full_partial_mismatch(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @exact_match("Hello")
        class Hello(LanguageSettings): pass

        @chain([Number10, Hello])
        class Number10ThenHello(LanguageSettings): pass

        iterator = StringParserIterator("10")
        with self.assertRaises(CannotParseException):
            Number10ThenHello(iterator)
        self.assertEqual("1", iterator.peek())

    def test_case_full_partial_mismatch2(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @exact_match("Hello")
        class Hello(LanguageSettings): pass

        @chain([Number10, Hello])
        class Number10ThenHello(LanguageSettings): pass

        iterator = StringParserIterator("10He")
        with self.assertRaises(CannotParseException):
            Number10ThenHello(iterator)
        self.assertEqual("1", iterator.peek())


if __name__ == '__main__':
    unittest.main()



import unittest

from minigrammar.iterators import StringParserIterator
from minigrammar.settings import LanguageSettings
from minigrammar.parsing import *
from minigrammar.exceptions import *


class ParsingRepeating(unittest.TestCase):

    def test_case_ok_packed(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @repeating(Number10, None, None, ',', True, False)
        class Numbers(LanguageSettings): pass

        iterator = StringParserIterator("10,10,10")
        Numbers(iterator)
        self.assertEqual(None, iterator.peek())

    def test_case_ok_well_far_apart(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @repeating(Number10, None, None, ',', True, False)
        class Numbers(LanguageSettings): pass

        iterator = StringParserIterator("10 , 10 , 10")
        Numbers(iterator)
        self.assertEqual(None, iterator.peek())

    def test_case_ok_mixed(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @repeating(Number10, None, None, ',', True, False)
        class Numbers(LanguageSettings): pass

        iterator = StringParserIterator(" 10 ,10,10")
        Numbers(iterator)
        self.assertEqual(None, iterator.peek())


    def test_case_mismatch_defaulted_to_sequence_of_zero_elems_ok(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @repeating(Number10, None, None, ',', True, False)
        class Numbers(LanguageSettings): pass

        iterator = StringParserIterator("XX, 10")
        Numbers(iterator)
        self.assertEqual("X", iterator.peek())

    def test_case_mismatch_defaulted_to_sequence_of_zero_elems_err(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @repeating(Number10, 5, None, ',', True, False)
        class Numbers(LanguageSettings): pass

        iterator = StringParserIterator("XX, 10")
        with self.assertRaises(CannotParseException):
            Numbers(iterator)
        self.assertEqual("X", iterator.peek())

    def test_case_too_many_elements(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @repeating(Number10, 1, 3, ',', True, False)
        class Numbers(LanguageSettings): pass

        iterator = StringParserIterator("10, 10, 10, 10")
        with self.assertRaises(CannotParseException):
            Numbers(iterator)
        self.assertEqual("1", iterator.peek())

    def test_case_trailing_used_partially_ok_parsed_prefix(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @repeating(Number10, None, None, ',', True, False)
        class Numbers(LanguageSettings): pass

        iterator = StringParserIterator("10, 10, 10, XX, ")
        Numbers(iterator)
        self.assertEqual("X", iterator.peek())

    def test_case_not_used_but_expected(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @repeating(Number10, None, None, ',', True, True)
        class Numbers(LanguageSettings): pass

        iterator = StringParserIterator("10, 10, 10, 10 ")
        with self.assertRaises(CannotParseException):
            Numbers(iterator)
        self.assertEqual("1", iterator.peek())


    def test_case_trailing_used_but_not_allowed(self):
        @exact_match("10")
        class Number10(LanguageSettings): pass

        @repeating(Number10, None, None, ',', False, False)
        class Numbers(LanguageSettings): pass

        iterator = StringParserIterator("10, 10, 10, 10, ")
        with self.assertRaises(CannotParseException):
            Numbers(iterator)
        self.assertEqual("1", iterator.peek())


if __name__ == '__main__':
    unittest.main()



class LanguageSettings:
    """
    This class contains language-specific configurations. The user is supposed to define a class that itself inherits
    from this, and overrides the needed configuration-methods. Then use that class as the base-class for all the grammar rules.
    The user-defined configuration-class should be in the same module of all the rules of the grammar.
    """
    context = {}

    def __init__(self):
        self.elems = []

    @classmethod
    def _get_id(cls):
        """
        Used by the parsing-logic injected by decorators to know their own fully-qualified rule-name.
        Should be inherited by a language-specific-configuration-class which itself inherits from LanguageSettings.
        :return: the fully-qualified name of this grammar-rule
        """
        return f"{cls.__module__}.{cls.__qualname__}"

    @classmethod
    def _fetch_parsing_rule_classname(cls, class_name: str | type) -> str:
        """
        This method is supposed to be aliased, and used to consistently refer to a given grammar-rule using
        just the name of the class. For this to work, such class must be defined within the same module.
        :param class_name: the name of a user-defined class, representing a grammar-rule, in the same module
        :return: the fully-qualified name of the given grammar-rule
        """
        if not isinstance(class_name, str): 
            return class_name._get_id()
        id_of_settings_class_without_last_part = cls._get_id()
        while id_of_settings_class_without_last_part[-1] != '.':
            id_of_settings_class_without_last_part = id_of_settings_class_without_last_part[:-1]
        return id_of_settings_class_without_last_part + class_name

    @classmethod
    def ignore_characters(cls, char: str) -> bool:
        """
        Used by the parsing-logic injected by decorators to know if a character should be ignored.
        Should be overridden by a language-specific configuration class which inherits from LanguageSettings.
        :param char: a single character to inspect
        :return: True if that character should be ignored, False otherwise
        """
        return char == ' ' or char == '\t' or char == '\n' or char == '\r'


from .exceptions import *

import re
import io

def _ignore_every_non_important_character(clazz, iterator_copy):
    while clazz.ignore_characters(iterator_copy.peek()):
        iterator_copy.advance()


def _keep_parsing_until_regex_match(pattern, string_buffer, iterator_copy):
    while iterator_copy.peek() is not None:
        if re.fullmatch(pattern, string_buffer.getvalue()):
            break
        string_buffer.write(iterator_copy.peek())
        iterator_copy.advance()


def _keep_parsing_until_regex_fail(pattern, string_buffer, iterator):
    while iterator.peek() is not None:
        if not re.fullmatch(pattern, string_buffer.getvalue() + iterator.peek()):
            break
        string_buffer.write(iterator.peek())
        iterator.advance()


def _attempt_parse_exact_match(keyword_or_symbol, iterator, fail_is_error):
    for char in keyword_or_symbol:
        if iterator.peek() != char:
            if fail_is_error:
                iterator.snapshot(0)
                raise CannotParseException(iterator)
            else:
                return None
        iterator.advance()
    return keyword_or_symbol


def _attempt_regex_match(pattern, iterator, string_buffer):
    if not re.match(pattern, string_buffer.getvalue()):
        iterator.snapshot(len(string_buffer.getvalue()))
        raise CannotParseException(iterator)


def _attempt_parse_rule_by_name(context, rule_name, iterator, fail_is_error):
    try:
        return context[rule_name](iterator)
    except CannotParseException:
        if fail_is_error:
            iterator.snapshot(0)
            raise CannotParseException(iterator)
        else:
            return None


def _expect_counter_within_bounds(counter, minimum, maximum, iterator):
    if minimum is not None and counter <= minimum:
        iterator.snapshot(0)
        raise CannotParseException(iterator)
    if maximum is not None and counter >= maximum:
        iterator.snapshot(0)
        raise CannotParseException(iterator)


def exact_match(keyword_or_symbol: str):
    """
    Used to decorate a user-defined class. That user defined class gets injected
    with all the necessary parsing-logic and becomes able to represent a grammar rule (e.g. a node in the AST).
    In particular, this decorator turns the class into a node representing a given keyword or a symbol.
    Such symbol gets stored in string format as the only element in the elems array, which is an instance field.
    :param keyword_or_symbol: The keyword or symbol to match.
    """
    def set_keyword_or_symbol_pattern_on_class(clazz):
        clazz.context[clazz._get_id()] = clazz
        def custom__init__(self, iterator_over_input_token_stream):
            iterator_copy = iterator_over_input_token_stream.clone()
            _ignore_every_non_important_character(clazz, iterator_copy)
            _attempt_parse_exact_match(keyword_or_symbol, iterator_copy, True)
            iterator_copy.synchronize_with_source()
            self.elems = [keyword_or_symbol]
        setattr(clazz, "__init__", custom__init__)
        return clazz
    return set_keyword_or_symbol_pattern_on_class


def regex_pattern(pattern: str):
    """
    Used to decorate a user-defined class. That user defined class gets injected
    with all the necessary parsing-logic and becomes able to represent a grammar rule (e.g. a node in the AST).
    In particular, this decorator turns the class into a node representing a given regex-pattern.
    The matching text will be stored in string format as the only element of the elems array, which
    is an instance field.
    :param pattern: The regex-pattern to match.
    """
    def set_regex_pattern_on_class(clazz):
        clazz.context[clazz._get_id()] = clazz
        def custom__init__(self, iterator_over_input_token_stream):
            iterator_copy = iterator_over_input_token_stream.clone()
            string_buffer = io.StringIO()
            _ignore_every_non_important_character(clazz, iterator_copy)
            _keep_parsing_until_regex_match(pattern, string_buffer, iterator_copy)
            _attempt_regex_match(pattern, iterator_copy, string_buffer)
            _keep_parsing_until_regex_fail(pattern, string_buffer, iterator_copy)
            iterator_copy.synchronize_with_source()
            self.elems = [string_buffer.getvalue()]
        setattr(clazz, "__init__", custom__init__)
        return clazz
    return set_regex_pattern_on_class


def repeating(rule_name: str, minimum: int | None, maximum: int | None, delimiter: str, allow_trailing: bool, enforce_trailing: bool):
    """
    Used to decorate a user-defined class. That user defined class gets injected
    with all the necessary parsing-logic and becomes able to represent a grammar rule (e.g. a node in the AST).
    In particular, this decorator turns the class into a node representing a given sequence of multiple grammar-rule
    occurrences of a grammar-rule separated by a delimiter. The matching text will be stored in string format as a
    sequence of elements of the elems array, which is an instance field.
    :param rule_name: The fully-qualified name of a decorated class
    :param minimum: The minimum number of times to repeat the rule (None defaults to 0)
    :param maximum: The maximum number of times to repeat the rule (None if unlimited)
    :param delimiter: The delimiter to use between occurrences of the rule
    :param allow_trailing: True means that trailing delimiters is allowed, False otherwise
    :param enforce_trailing: True means that trailing delimiters is enforced, False otherwise
    """
    def set_multiple_repeating_rules_on_class(clazz):
        clazz.context[clazz._get_id()] = clazz
        def custom__init__(self, iterator_over_input_token_stream):
            iterator_copy = iterator_over_input_token_stream.clone()
            _ignore_every_non_important_character(clazz, iterator_copy)
            unaliased_name = clazz._fetch_parsing_rule_classname(rule_name)
            parsed_element = _attempt_parse_rule_by_name(clazz.context, unaliased_name, iterator_copy, minimum is not None and minimum <= 0)
            counter = 0 if parsed_element is None else 1
            extracted_delimiter = delimiter
            self.elems = [parsed_element]
            while parsed_element is not None and extracted_delimiter is not None:
                _ignore_every_non_important_character(clazz, iterator_copy)
                extracted_delimiter = _attempt_parse_exact_match(delimiter, iterator_copy, enforce_trailing)
                if extracted_delimiter is not None:
                    _ignore_every_non_important_character(clazz, iterator_copy)
                    parsed_element = _attempt_parse_rule_by_name(clazz.context, unaliased_name, iterator_copy, not allow_trailing)
                    self.elems.append(parsed_element)
                counter += 1
            _expect_counter_within_bounds(counter, minimum, maximum, iterator_copy)
            iterator_copy.synchronize_with_source()
        setattr(clazz, "__init__", custom__init__)
        return clazz
    return set_multiple_repeating_rules_on_class


def chain(rule_names: list[str]):
    """
    Used to decorate a user-defined class. That user defined class gets injected
    with all the necessary parsing-logic and becomes able to represent a grammar rule (e.g. a node in the AST).
    In particular, this decorator turns the class into a node representing a given list of multiple different grammar-rules
    :param rule_names: A list of fully-qualified decorated class names
    """
    def set_multiple_sequential_rules_on_class(clazz):
        clazz.context[clazz._get_id()] = clazz
        def custom__init__(self, iterator_over_input_token_stream):
            iterator_copy = iterator_over_input_token_stream.clone()
            _ignore_every_non_important_character(clazz, iterator_copy)
            self.elems = []
            for r_name in rule_names:
                unaliased_name = clazz._fetch_parsing_rule_classname(r_name)
                parsed_element = _attempt_parse_rule_by_name(clazz.context, unaliased_name, iterator_copy, True)
                self.elems.append(parsed_element)
            iterator_copy.synchronize_with_source()
        setattr(clazz, "__init__", custom__init__)
        return clazz
    return set_multiple_sequential_rules_on_class


def either(rule_names: list[str]):
    """
    Used to decorate a user-defined class. That user defined class gets injected
    with all the necessary parsing-logic and becomes able to represent a grammar rule (e.g. a node in the AST).
    In particular, this decorator turns the class into a node representing one single grammar-rule within a list
    of possible grammar-rules.
    :param rule_names: A list of fully-qualified decorated class names
    """
    def set_mutually_exclusive_rules_on_class(clazz):
        clazz.context[clazz._get_id()] = clazz
        def custom__init__(self, iterator_over_input_token_stream):
            iterator_copy = iterator_over_input_token_stream.clone()
            _ignore_every_non_important_character(clazz, iterator_copy)
            for r_name in rule_names:
                unaliased_name = clazz._fetch_parsing_rule_classname(r_name)
                self._elem = _attempt_parse_rule_by_name(clazz.context, unaliased_name, iterator_copy, False)
                if self._elem is not None:
                    self.elems = [self._elem]
                    iterator_copy.synchronize_with_source()
                    return
            raise CannotParseException(iterator_copy)
        setattr(clazz, "__init__", custom__init__)
        return clazz
    return set_mutually_exclusive_rules_on_class


from minigrammar import *


class MathSettings(LanguageSettings):
    @classmethod
    def ignore_characters(cls, char):
        return char == ' ' or char == '\t' or char == '\r' or char == '\n'


@repeating("Addend", 1, None, '+', False, False)
class Expression(MathSettings):
    def __repr__(self):
        string = self.elems[0].__repr__()
        for elem in self.elems[1:]:
            string += " + " + repr(elem)
        return " { " + string + " } "


@repeating("Factor", 1, None, '*', False, False)
class Addend(MathSettings):
    def __repr__(self):
        string = self.elems[0].__repr__()
        for elem in self.elems[1:]:
            string += " * " + repr(elem)
        return " { " + string + " } "


@either(["Number", "Variable", "WrappedExpression"])
class Factor(MathSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


@regex_pattern(r'^\d+$')
class Number(MathSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


@regex_pattern(r'^[a-z]$')
class Variable(MathSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


@chain(["OpenParen", "Expression", "ClosedParen"])
class WrappedExpression(MathSettings):
    def __repr__(self):
        return " ( " + self.elems[1].__repr__() + " ) "


@exact_match("(")
class OpenParen(MathSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


@exact_match(")")
class ClosedParen(MathSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


if __name__ == "__main__":
    iterator = StringParserIterator("x*(a+b)")
    expr = Expression(iterator)
    print(expr)



import sys
from minigrammar import *


class JsonSettings(LanguageSettings):
    @classmethod
    def ignore_characters(cls, char):
        return char == ' ' or char == '\t' or char == '\r' or char == '\n'


@either(["Array", "Value", "Object"])
class Json(JsonSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


@chain(["OpenSquareBracket", "MultipleArrayElems", "ClosedSquareBracket"])
class Array(JsonSettings):
    def __repr__(self):
        return "[" + self.elems[1].__repr__() + "]"


@chain(["OpenCurlyBracket", "MultipleObjectFields", "ClosedCurlyBracket"])
class Object(JsonSettings):
    def __repr__(self):
        return "{" + self.elems[1].__repr__() + "}"


@repeating("Json", None, None, ',', False, False)
class MultipleArrayElems(JsonSettings):
    def __repr__(self):
        return self.elems.__repr__()[1:-1]


@repeating("ObjectField", None, None, ',', False, False)
class MultipleObjectFields(JsonSettings):
    def __repr__(self):
        return self.elems.__repr__()[1:-1]


@chain(["StringLiteral", "Colon", "Json"])
class ObjectField(JsonSettings):
    def __repr__(self):
        return self.elems[0].__repr__() + " : " + self.elems[2].__repr__()


@either(["StringLiteral", "FloatingPointLiteral", "IntegerLiteral", "BooleanLiteral", "NullLiteral"])
class Value(JsonSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


@regex_pattern(r'^"(?:[^"\\]|\\.)*"$')
class StringLiteral(JsonSettings):
    def __repr__(self):
        return self.elems[0].__repr__()[1:-1]


@exact_match("null")
class NullLiteral(JsonSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


@regex_pattern(r'^\d+$')
class IntegerLiteral(JsonSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


@regex_pattern(r'^\d*\.\d+$')
class FloatingPointLiteral(JsonSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


@regex_pattern(r'true|false')
class BooleanLiteral(JsonSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


@exact_match("[")
class OpenSquareBracket(JsonSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


@exact_match("]")
class ClosedSquareBracket(JsonSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


@exact_match("{")
class OpenCurlyBracket(JsonSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


@exact_match("}")
class ClosedCurlyBracket(JsonSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


@exact_match(":")
class Colon(JsonSettings):
    def __repr__(self):
        return self.elems[0].__repr__()


if __name__ == "__main__":
    with open(sys.argv[1], 'r') as file:
        iterator = FileParserIterator(file)
        json = Json(iterator)
        print(json)


from collections.abc import Callable
from typing import TextIO


class ParserIterator:
    """
    It's used as the one and only parameter for constructors (__init__) of AST-classes where parsing a file is needed.
    This class can be used as a python-iterator (e.g. using next on it) to iterate over characters of an indexable
    source of characters representing the text to be parsed
    """

    def __init__(self, get_at_index: Callable[[int], str]):
        self._index = 0
        self._char_pos = 1
        self._line_number = 1
        self._get_at_index = get_at_index
        self._generator = None
        self._snapshots = []

    def __iter__(self):
        return self

    def __next__(self):
        current_value = self.peek()
        if current_value is None:
            raise StopIteration
        self.advance()
        return current_value

    def clone(self) -> 'ParserIterator':
        """
        :return: an independent clone of this iterator, with its own state
        """
        cloned = ParserIterator(self._get_at_index)
        cloned._snapshots = self._snapshots
        cloned._char_pos = self._char_pos
        cloned._line_number = self._line_number
        cloned._index = self._index
        cloned._generator = self
        return cloned

    def advance(self) -> None:
        """
        Advances the iterator by one character. It doesn't return any value,
        to query for the current value of the iterator use the peek() method
        """
        self._index += 1
        self._char_pos += 1
        if self._get_at_index(self._index) == "\n":
            self._char_pos = 1
            self._line_number += 1

    def advance_by(self, n: int) -> None:
        """
        Advances the iterator by n character. It doesn't return any value,
        to query for the current value of the iterator use the peek() method
        :param n: how many times to advance the iterator
        """
        for _ in range(n):
            self.advance()

    def peek(self) -> str | None:
        """
        Used to see the current value of the iterator without advancing it.
        :return: a string of length 1 corresponding to the current value of the iterator or None if end-of-file is reached.
        """
        current_char = self._get_at_index(self._index)
        if current_char == "":
            return None
        return current_char

    def synchronize_with_source(self) -> None:
        """
        When called on an iterator that has been created using the clone method on another iterator,
        it will update the state of the original iterator to match the state of this iterator. If the
        iterator has not been created as a clone, it does nothing.
        """
        if self._generator is not None:
            self._generator._index = self._index
            self._generator._char_pos = self._char_pos
            self._generator._line_number = self._line_number

    def snapshot(self, offset : int = 0):
        """
        Used to save the current coordinates of the iterator and later recover the location of an error 
        even in cases where the original iterator has been advanced or restored to a previous state.
        """
        self._snapshots.append((self._line_number, self._char_pos - offset))

    def inspect_for_errors(self) -> tuple[int, int]:
        """
        Used to get the furthest snapshot saved so far, which corresponds to the location of the most 
        advanced error. This always returns something even if no exception were raised during parsing.
        You should only care about inspecting for errors if you catched an exception in a try/catch block.
        :return: a tuple (line_number, char_pos) corresponding to the furthest snapshot saved so far.
        """
        if len(self._snapshots) == 0:
            return (1, 1)
        return max(self._snapshots, key=lambda x: (x[0], x[1]))



class FileParserIterator(ParserIterator):
    """
    It's used as the one and only parameter for constructors (__init__) of AST-classes where parsing
    a file is needed. This class can be used as a python-iterator (e.g. using next on it) to iterate
    over characters of a file. It must be constructed from a file already opened in read mode,
    corresponding to the file to parse.
    """

    def __init__(self, source_file: TextIO):
        def get_at_index(index: int):
            source_file.seek(index)
            current_char = source_file.read(1)
            source_file.seek(index)
            return current_char

        super().__init__(get_at_index)


class StringParserIterator(ParserIterator):
    """
    It's used as the one and only parameter for constructors (__init__) of AST-classes where parsing a
    file is needed. This class can be used as a python-iterator (e.g. using next on it) to iterate
    over characters of a string. It must be constructed from a string containing text to parse.
    """

    def __init__(self, source_text: str):
        def get_at_index(index: int):
            if index >= len(source_text):
                return None
            else:
                return source_text[index]

        super().__init__(get_at_index)


class CannotParseException(Exception):
    def __init__(self, iterator):
        line_number, char_position = iterator.inspect_for_errors()
        super().__init__(f"Can't parse the given entity at line {line_number}, character {char_position}")
